#include <stdio.h>
#include <string.h>
#include "neocd.h"
#include "cache.h"

#define printf	pspDebugScreenPrintf

typedef struct {
  char name[15];
  int offset, len, crc;
} file_entry;

static file_entry *cache[7];
static int used[7];
static int max[7];

int file_cache(char *filename, int offset, int size,int type) {
  int use;
  char *s = strrchr(filename,'/');
  if (s)
    filename = s+1; // no path here !

  if (cache[type]) {
    file_entry *list = cache[type];
    int n = 0;
    while (n < used[type] && list[n].offset < offset)
      n++;
    if (n == used[type]) { // reached the end of list
      use = used[type]++;
      if (used[type] >= max[type]) {
	// I use >= on purpose because I want more room for an eventual insertion
	// see memmove further
	max[type]+=32;
	cache[type] = realloc(cache[type],sizeof(file_entry)*max[type]);
      }
    } else { // found an offset >= this one
      int end;
      if (list[n].offset == offset && list[n].len == size &&
	  !strcmp(list[n].name,filename)) {
	// already have it !
	int old_crc = list[n].crc;
	// printf("!!! already have it !!!\n");
	cache_set_crc(offset,size,type);
	if (list[n].crc != old_crc) {
	  // printf("... but crcs differ ! ;-(\n");
	  return 0;
	}
	return 1;
      }
      end = offset+size;
      // First : skip eventually biger files loaded here (we overwrite the beg so we
      // should be loaded after them
      while (list[n].offset == offset && list[n].offset+list[n].len > end)
	n++;

      // remove those which are overwriten (if there are some)
      while (n < used[type] && list[n].offset+list[n].len <= end) {
	memmove(&list[n],&list[n+1],sizeof(file_entry)*(used[type]-(n+1)));
	used[type]--;
      }
      // otherwise insertion needed at this place
      if (n < used[type])
	memmove(&list[n+1],&list[n],(used[type]-n)*sizeof(file_entry));
      use = n;
      used[type]++;
      if (used[type] >= max[type]) {
	max[type]+=32;
	cache[type] = realloc(cache[type],sizeof(file_entry)*max[type]);
      }
    }
  } else { // first allocation
    if (!max[type])
      max[type] = 32;
    cache[type] = malloc(sizeof(file_entry)*max[type]);
    used[type] = 1;
    use = 0;
  }
  strcpy(cache[type][use].name,filename);
  cache[type][use].offset = offset;
  cache[type][use].len = size;
  return 0; // not already cached, new file
}

int get_used(int type) {
  if (type < 7)
    return used[type];
  return -1;
}

void clear_file_cache() {
  int n;
  for (n=0; n<7; n++) {
    if (cache[n]) {
      free(cache[n]);
      cache[n] = NULL;
      used[n] = 0;
      max[n] = 32;
    }
  }
}

void cache_set_crc(int offset,int size,int type) {
  /* The problem of neocd is that everything is loaded in ram, which means that it can
     be changed dynamically after it's loaded. Most of the time this is only changed
     by accessing the upload area so we could probably only test the upload area.
     But the 68k and z80 prg memory, can be changed directly or with some patches
     for the z80. So the best way is to add a crc */
  if (cache[type]) {
    file_entry *list = cache[type];
    int n = 0;
    char *src=NULL;
    unsigned short crc;

    while (n < used[type] && list[n].offset < offset)
      n++;

    /* Normally this function is called AFTER file_cache, which means that this loop
       will always find the correct file in the cache, we just need to compute the crc
       here */

    switch(type) {
    case PRG_TYPE: src= neogeo_prg_memory + offset; break;
    case Z80_TYPE: src= ((char*)subcpu_memspace) + offset; break;
    case FIX_TYPE: src= neogeo_fix_memory + offset; break;
    case SPR_TYPE: src= neogeo_spr_memory + offset; break;
    case PCM_TYPE: src= neogeo_pcm_memory + offset; break;
    }

    crc = 0;
    /* If the size is not a multiple of 4, then we loose precision, a few bytes are
       ignored at the end of the file. Oh well it should be enough just to check
       if an area has changed anyway... ! */
    do {
      crc ^= *((int *)src);
      size -= 4;
      src += 4;
    } while (size>=4);
    list[n].crc = crc;
  }
}

