/********************************************
*   NeoCD Memory Mapping (C version)        *
*********************************************
* Fosters(2001,2003)                        *
********************************************/

//#include <SDL/SDL.h>
//#include <SDL/SDL_endian.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "neocd.h"
#include "psp/blitter.h"
#include "cdrom/cdrom.h"

#ifdef DEBUG_MEMORY
extern int trazando;
#endif

extern int nb_interlace;

char mem_str[256];

extern unsigned short    *video_paletteram_ng;

#if 0
    #define logaccess(...) printf(__VA_ARGS__)
#else
    #define logaccess(...)
#endif

#define RASTER_COUNTER_START 0x1f0	/* value assumed right after vblank */
#define RASTER_COUNTER_RELOAD 0x0f8	/* value assumed after 0x1ff */
#define RASTER_LINE_RELOAD (0x200 - RASTER_COUNTER_START)

/*** Globals ***************************************************************************************/
int watchdog_counter=-1;
int memcard_write=0;

void   initialize_memmap(void);
unsigned int m68k_read_memory_8(unsigned int address);
unsigned int m68k_read_memory_16(unsigned int address);
unsigned int m68k_read_memory_32(unsigned int address);
void m68k_write_memory_8(unsigned int address, unsigned int value);
void m68k_write_memory_16(unsigned int address, unsigned int value);
void m68k_write_memory_32(unsigned int address, unsigned int value);


/***Helper Functions********************************************************************************/
static int    read_upload(int);
static void   write_upload(int, int);
static void   write_upload_word(int, int);

static int    cpu_readvidreg(int);
static void   cpu_writevidreg(int, int);
static void   cpu_writeswitch(int, int);
static int    cpu_readcoin(int);

static void   write_memcard(int,int);

void cdda_control(void);

static void   watchdog_reset_w(void);
static char game_vectors[0x80];


/***************************************************************************************************/
void initialize_memmap(void) {
  memcpy(game_vectors,neogeo_prg_memory,0x80);
  return;
}


/***************************************************************************************************/

void	swab( const void* src1, const void* src2, int isize)
{
	char*	ptr1;
	char*	ptr2;
	char	tmp;
	int		ic1;


	ptr1 = (char*)src1;
	ptr2 = (char*)src2;
	for ( ic1=0 ; ic1<isize-1 ; ic1+=2){
		tmp = ptr1[ic1+0];
		ptr2[ic1+0] = ptr1[ic1+1];
		ptr2[ic1+1] = tmp;
	}


}

unsigned int  FASTCALL m68k_read_memory_8(const unsigned int offset) {
    if(offset<0x200000)
        return neogeo_prg_memory[offset^1];

    switch(offset>>16)
    {
        case    0x30:    return read_player1();
    case 0x33:
    case    0x32:    return cpu_readcoin(offset);
        case    0x34:    return read_player2();
        case    0x38:    return read_pl12_startsel();
        case    0x80:    if(offset&0x01) return neogeo_memorycard[(offset&0x3fff)>>1];
                         else return -1;

        /* BIOS ROM */
        case    0xc0:
        case    0xc1:
        case    0xc2:
        case    0xc3:
        case    0xc4:
        case    0xc5:
        case    0xc6:
        case    0xc7:    return neogeo_rom_memory[(offset^1)&0x0fffff];

        case    0xe0:
        case    0xe1:
        case    0xe2:
        case    0xe3:    return read_upload(offset&0xfffff);

        default:
          //  logaccess("m68k_read_memory_8(0x%x) PC=%x\n",offset,m68k_get_reg(NULL,M68K_REG_PC));
	  if (offset & 0xff000000)
	    /* Ironclad at least needs this. Maybe because it's still more or
	       less a prototype ? Anyway moving this stuff at the end speeds
	       noticably things up for normal well writen games */
	    return m68k_read_memory_8(offset & 0xffffff);
    }
    return 0xff;
}

/***************************************************************************************************/
unsigned int  FASTCALL m68k_read_memory_16(const unsigned int offset) {
    unsigned int off16;
    if(offset<0x200000)
        return *((u16*)&neogeo_prg_memory[offset]);
    off16 = offset >> 16;

    switch(off16)
    {
    case 0x33:
    case    0x32:    return cpu_readcoin(offset);
        case    0x3c:    return cpu_readvidreg(offset);
        case    0x80:    return 0xff00|neogeo_memorycard[(offset&0x3fff)>>1];

        /* BIOS ROM */
        case    0xc0:
        case    0xc1:
        case    0xc2:
        case    0xc3:
        case    0xc4:
        case    0xc5:
        case    0xc6:
        case    0xc7:    return *(u16*)&neogeo_rom_memory[offset&0x0fffff];

		case	0xff:	 break;


        default:
          //  logaccess("m68k_read_memory_16(0x%x) PC=%x\n",offset,m68k_get_reg(NULL,M68K_REG_PPC));
	  if (off16 >= 0x40 && off16 <= 0x7f) {
	    // mirrored palette
	    return video_paletteram_ng[(offset&0x1fff)>>1];
	  } else if (offset & 0xff000000)
	    /* Ironclad at least needs this. Maybe because it's still more or
	       less a prototype ? Anyway moving this stuff at the end speeds
	       noticably things up for normal well writen games */
	    return m68k_read_memory_16(offset & 0xffffff);
    }
    return 0xffff;
}


/***************************************************************************************************/
unsigned int  FASTCALL m68k_read_memory_32f(const unsigned int off) {
#ifdef DEBUG_MEMORY
if (trazando) printf("m68k_read_memory_32f(0x%X)\n",off); fflush(stdout);
#endif
    unsigned int offset = off & 0xffffff;
    unsigned int data;
    data=m68k_read_memory_16(offset)<<16;
    data|=m68k_read_memory_16(offset+2);
    return data;
}

unsigned int  m68k_read_memory_32(unsigned int offset) {
#ifdef DEBUG_MEMORY
if (trazando) printf("m68k_read_memory_32(0x%X)\n",offset); fflush(stdout);
#endif
    return m68k_read_memory_32f(offset);
}



/***************************************************************************************************/
void FASTCALL m68k_write_memory_8(const unsigned int offset, unsigned int data) {
    int temp;
    if(offset<0x200000) {
        neogeo_prg_memory[offset^1]=(char)data;
	if ((offset == ((0x108000+30455)^1)) && data <= 7) {
	  neogeo_do_cdda(data,neogeo_prg_memory[(0x108000+30456)^1]);
	} else if (offset == 0x10f651) { // test end loading of files
	  neogeo_prg_memory[0x10f651] = data; // duplicate write
	}
        return;
    }

    switch(offset>>16)
    {
        case    0x30:    watchdog_reset_w(); break;
        case    0x32:

	  if (neogeo_sound_volume) {
	    sound_code=data;
	    pending_command=1;
	    mz80nmi();
	    ExitOnEI = 1;
	    mz80exec(60000);
	  }

	  break;
		case	0x38:	break;
        case    0x3a:    cpu_writeswitch(offset, data); break;
        case    0x3c:    temp=cpu_readvidreg(offset);
            if(offset&0x01) cpu_writevidreg(offset, (temp&0xff)|(data<<8));
            else cpu_writevidreg(offset, (temp&0xff00)|data);
            break;
        case    0x80:    if(offset&0x01) write_memcard(offset,data); break;


        /* upload */
        case    0xe0:
        case    0xe1:
        case    0xe2:
        case    0xe3:   write_upload(offset&0xfffff,data); break;
		/* cdrom */
		case	0xff:	break;

        default:
	  if (offset & 0xff000000)
	    return m68k_write_memory_8(offset & 0xffffff,data);
       //     logaccess("m68k_write_memory_8(0x%x,0x%x) PC=%x\n",offset,data,m68k_get_reg(NULL,M68K_REG_PC));
            break;
    }
}


/***************************************************************************************************/
void FASTCALL m68k_write_memory_16(const unsigned int offset, unsigned int data) {
    unsigned int off16;

    if(offset<0x200000) {
        *(u16*)&neogeo_prg_memory[offset]=(u16)data;
        return;
    }
    off16 = offset >> 16;

    switch(off16)
      {
      case	0x2f:	break;
      case    0x32:

	if (neogeo_sound_volume) {
	  sound_code=(data >> 8);
	  pending_command=1;
	  mz80nmi();
	  ExitOnEI = 1;
	  mz80exec(60000);
	}

	break;
      case    0x3a:    cpu_writeswitch(offset, data); break;
      case    0x3c:    cpu_writevidreg(offset, data); break;
      case    0x80:    write_memcard(offset,data); break;

        /* upload */
      case    0xe0:
      case    0xe1:
      case    0xe2:
      case    0xe3:    write_upload_word(offset&0xfffff,data); break;
      case	0xff:
	if (offset == 0xff0002) {
	  if (data == 0x550) {
	    if (neogeo_prg_memory[0x115A06 ^ 1]>32 && neogeo_prg_memory[0x115A06 ^ 1] < 127) {
	      neogeo_cdrom_load_files(&neogeo_prg_memory[0x115a06]);
	    }
	  }
	}
	break;

      default:
	if (off16 >= 0x40 && off16 <= 0x7f) {
	  // mirrored palette
	  int offset2 =(offset&0x1fff)>>1;
	  data  &=0x7fff;
	  blit_fix_resetcache();blit_spr_resetcache();
	  video_paletteram_ng[offset2]=(u16)data;
	  video_paletteram_pc[offset2]=video_color_lut[data]|(offset2&15?0:0x8000);
	} else if (offset & 0xff000000)
	  return m68k_write_memory_16(offset & 0xffffff,data);

	//  logaccess("m68k_write_memory_16(0x%x,0x%x) PC=%x\n",offset,data,m68k_get_reg(NULL,M68K_REG_PC));
      }
}


/***************************************************************************************************/
void FASTCALL m68k_write_memory_32(const unsigned int off, unsigned int data) {
#ifdef DEBUG_MEMORY
if (trazando) printf("m68k_write_memory_32(0x%X,0x%X)\n",off,data); fflush(stdout);
#endif
      unsigned int offset = off & 0xffffff;
	unsigned int word1=(data>>16)&0xffff;
	unsigned int word2=data&0xffff;
	m68k_write_memory_16(offset,word1);
	m68k_write_memory_16(offset+2,word2);
}


/***************************************************************************************************/
static int    cpu_readvidreg(int offset)
{
  switch(offset)
    {
    case    0x3c0000:    return *(u16*)&video_vidram[video_pointer<<1]; break;
    case    0x3c0002:    return *(u16*)&video_vidram[video_pointer<<1]; break;
    case    0x3c0004:    return video_modulo; break;
    case    0x3c0006:    // printf("read rastercounter %d frame_counter %d\n",rastercounter,neogeo_frame_counter);
      /* Determine the rastercounter based on the number of executed
	 cycles */
      {
	if (!raster_frame) {
	  // scanline is not available, find it !
/* 	  int cycles = ((Z80_VBL_CYCLES*3) - (C68K.CycleIO)); */
/* 	  scanline = cycles*264.0/(Z80_VBL_CYCLES*3); */
/* 	  printf("scanline from cycles %d\n",scanline); */
	  if (scanline == 263)
	    scanline = 16;
	  else
	    scanline = 263;
	}
	int rastercounter;

	if (scanline < RASTER_LINE_RELOAD)
	  rastercounter = RASTER_COUNTER_START + scanline;
	else
	  rastercounter = RASTER_COUNTER_RELOAD + scanline - RASTER_LINE_RELOAD;
	// printf("read rastercounter %d neogeo_frame_counter %d scanline %d\n",rastercounter,neogeo_frame_counter,scanline);
	return ((rastercounter<<7)&0xff80)|
	  (neogeo_frame_counter&7);
      }
    default:
      return 0;
      break;
    }
}
#if 0
// This is a crazy situation : this version which should work in more cases
// than the previous one breaks many games which use the raster interrupt
// like aof2 or viewpoint. Crazy, simply crazy. Use the first version and
// don't bother !!!
static int    cpu_readvidreg(int offset)
{
  offset &= 0xe;
  switch(offset)
    {
        case    0:    return *(u16*)&video_vidram[video_pointer<<1]; break;
        case    2:    return *(u16*)&video_vidram[video_pointer<<1]; break;
        case    4:    return video_modulo; break;
        case    6:
	  {
	    int rastercounter;
	    if (!raster_frame) { // must find scanline
	      /* The raster interrupt is not active.
		 In this case, just return the last scanline, which is
		 enough to get out of a waiting loop quickly... */
	      if (scanline == 263)
		scanline = 16;
	      else
		scanline = 263;
	    }
/* 	      int cycles = C68k_Get_CycleDone(&C68K); */
/* 	      scanline = cycles*264.0/NEO4ALL_Z80_NORMAL_CYCLES; */
	    if (scanline < RASTER_LINE_RELOAD)
	      rastercounter = RASTER_COUNTER_START + scanline;
	    else
	      rastercounter = RASTER_COUNTER_RELOAD + scanline - RASTER_LINE_RELOAD;
	    return ((rastercounter<<7)&0xff80)|
	    (neogeo_frame_counter&7);
	  }
        case    8:    return *(u16*)&video_vidram[video_pointer<<1]; break;
        case    0xa:    return *(u16*)&video_vidram[video_pointer<<1]; break;

        default:
	      printf("cpu_readvidreg(0x%x)\n",offset);
            break;
    }
  return 0;
}
#endif

/***************************************************************************************************/
static void neo_irq2pos_w(int offset, u16 data)
{
  /* This is entierly from mame, so thanks to them ! */
	if (offset)
		irq2pos = (irq2pos & 0xffff0000) | data;
	else
		irq2pos = (irq2pos & 0x0000ffff) | (data << 16);

	if (irq2control & IRQ2CTRL_LOAD_RELATIVE)
	{
//      int line = (irq2pos_value + 3) / 0x180; /* ridhero gives 0x17d */
	  int line = (irq2pos + 0x3b) / 0x180;	/* turfmast goes as low as 0x145 */

	  if (scanline == 263)
	    irq2start = line;
	  else
	    irq2start = (scanline + line);
	}
}

static void    cpu_writevidreg(int offset, int data)
{
  offset &= 0xe;
    switch(offset)
    {
        case    0:    video_pointer=(u16)data; break;
        case    2:    *(u16*)&video_vidram[video_pointer<<1]=(u16)data;
	  video_pointer+=video_modulo;
	  // video_pointer=(video_pointer & 0x8000) | ((video_pointer +video_modulo) & 0x7fff);
	  //blit_spr_resetcache();
	  break;
        case    4:    video_modulo=(s16)data; break;

        case    6:
	  neogeo_frame_counter_speed=((data>>8)&0xff)+1;
	  irq2control = data & 0xff;
	  break;

    case    8:   neo_irq2pos_w(0,data); /* IRQ position */    break;
    case    0xa: neo_irq2pos_w(1,data); /* IRQ position */    break;
    case    0xc:    /* IRQ acknowledge */
      if (data & 4) vblank_int = 0;
      if (data & 2) scanline_int = 0;
      break;

    default:
        //    logaccess("cpu_writevidreg(0x%x,0x%x) PC=%x\n",offset,data,m68k_get_reg(NULL,M68K_REG_PC));
            break;
    }
}

/***************************************************************************************************/

static void     neogeo_setpalbank0 (void) {
#ifdef DEBUG_MEMORY
if (trazando) puts("neogeo_setpalbank0()"); fflush(stdout);
#endif
    video_paletteram_ng=video_palette_bank0_ng;
    video_paletteram_pc=video_palette_bank0_pc;

}


static void     neogeo_setpalbank1 (void) {
#ifdef DEBUG_MEMORY
if (trazando) puts("neogeo_setpalbank1()"); fflush(stdout);
#endif
    video_paletteram_ng=video_palette_bank1_ng;
    video_paletteram_pc=video_palette_bank1_pc;
}


static void    neogeo_select_bios_vectors (void) {
#ifdef DEBUG_MEMORY
if (trazando) puts("neogeo_select_bios_vectors()"); fflush(stdout);
#endif
    memcpy(neogeo_prg_memory, neogeo_rom_memory, 0x80);
}


static void    neogeo_select_game_vectors (void) {
#ifdef DEBUG_MEMORY
if (trazando) puts("neogeo_select_game_vectors()"); fflush(stdout);
#endif
// memcpy(neogeo_prg_memory,game_vectors,0x80);
    logaccess("write game vectors stub\n");
}


static void    cpu_writeswitch(int offset, int data)
{
#ifdef DEBUG_MEMORY
if (trazando) printf("cpu_writeswitch(0x%X,0x%X)\n",offset,data); fflush(stdout);
#endif
 offset &= 0x1e;
    switch(offset)
    {
        case 0: /* NOP */ break;

        case 2: neogeo_select_bios_vectors(); break;

        case 0xe: neogeo_setpalbank1(); break;

        case 0x10: /* NOP */ break;

        case 0x12: neogeo_select_game_vectors(); break;

        case 0x1e: neogeo_setpalbank0(); break;

        default:
       //     logaccess("cpu_writeswitch(0x%x,0x%x) PC=%x\n",offset,data,m68k_get_reg(NULL,M68K_REG_PC));
            break;
    }
}

/***************************************************************************************************/

void neogeo_sound_irq(int irq)
{
#ifdef DEBUG_MEMORY
if (trazando) printf("neogeo_sound_irq(0x%X)\n",irq); fflush(stdout);
#endif

    if (neogeo_sound_volume)
    {
    	if (irq) {
        	//logaccess("neogeo_sound_irq %d\n",irq);
        	mz80int(0);
    	} else {
		mz80ClearPendingInterrupt();
        	//mz80ClearPendingInterrupt();
        	//logaccess("neogeo_sound_end %d\n",irq);
	}
    }

}



static int cpu_readcoin(int addr)
{
#ifdef DEBUG_MEMORY
if (trazando) printf("cpu_readcoin(0x%X)\n",addr); fflush(stdout);
#endif
    addr &= 0xFFFF;
    if (addr & 0x1) {
        int coinflip = pd4990a_testbit_r();
        int databit = pd4990a_databit_r();
        return 0xff ^ (coinflip << 6) ^ (databit << 7);
    }
    {
      int res = result_code;

      if (neogeo_sound_volume)
	{
        	if (pending_command)
           		 res &= 0x7f;
        }

        return res;
    }
}


static void watchdog_reset_w (void)
{
#ifdef DEBUG_MEMORY
if (trazando) puts("watchdog_reset_w()"); fflush(stdout);
#endif
    //if (watchdog_counter == -1) logaccess("Watchdog Armed!\n");
    watchdog_counter=3 * 60;  /* 3s * 60 fps */
}

static int read_upload(int offset) {
    int zone = m68k_read_memory_8(0x10FEDA);
   // int bank = m68k_read_memory_8(0x10FEDB);

    switch (zone) {
    case 0x00:  // /* 68000 */          return neogeo_prg_memory[offset^1];
      // return subcpu_memspace[(offset>>1)^1];

    case 0x01: /* Z80 */
      if (broken_upload_area) // only for neogeo special for now
	return 0xff;
      return subcpu_memspace[offset>>1];
    case 0x11: /* FIX */
      return neogeo_fix_memory[offset>>1];
    default:
           //sprintf(mem_str,"read_upload unimplemented zone %x\n",zone);
           //debug_log(mem_str);
	  // This is probably not the z80, or at least not this way.
	  // try master of monsters 2 for a clear demonstration that
	  // it can't possibly work this way
            return -1;
    }
}


static void write_upload(int offset, int data) {
    int zone = m68k_read_memory_8(0x10FEDA);
    /* int bank = m68k_read_memory_8(0x10FEDB); */
    // This write is really used at least by metal slug to update the
    // fix area when choosing "Exit" from the main screen. (cd screen)

    switch (zone) {
        case 0x00: /* 68000 */
	  /* It this really 68k write ?
	     anyway, this breaks breakers, just start to play and hit
	     the opponent (or be hit), it hangs the game if this write
	     is authorized */
	  // neogeo_prg_memory[offset^1]=(char)data;
	  // offset >>=1;
	  // if (offset >= 0xf800) // ???
	  // subcpu_memspace[(offset)^1]=(char)data;
	  break;
        case 0x01: /* Z80 */
	  /* The writes must be allowed to the *whole* z80 memory.
	     It's used by at least samurai spirits rpg to indicate
	     the locations of the samples */
	  subcpu_memspace[offset>>1]=data;
	  break;
        case 0x11: /* FIX */
            neogeo_fix_memory[offset>>1]=(char)data;

            blit_fix_resetcache();
            break;
        default:
         //sprintf(mem_str,"write_upload unimplemented zone %x\n",zone);
         //debug_log(mem_str);
			break;
    }
}


static void write_upload_word(int offset, int data) {
    int zone = m68k_read_memory_8(0x10FEDA);
    int bank = m68k_read_memory_8(0x10FEDB);
    int offset2;
    char *dest;
    char sprbuffer[4];

  	data&=0xffff;

    switch (zone) {
        case 0x12: /* SPR */

            offset2=offset & ~0x02;

            offset2+=(bank<<20);

            if((offset2&0x7f)<64)
               offset2=(offset2&0xfff80)+((offset2&0x7f)<<1)+4;
            else
               offset2=(offset2&0xfff80)+((offset2&0x7f)<<1)-128;

            dest=&neogeo_spr_memory[offset2];
            blit_spr_resetcache();

            if (offset & 0x02) {
               /* second word */
			   *(u16*)(&dest[2])=(u16)data;
			   /* reformat sprite data */
               swab(dest, sprbuffer, sizeof(sprbuffer));
               extract8(sprbuffer, dest);
			} else {
			   /* first word */
			   *(u16*)(&dest[0])=(u16)data;
            }
            break;

        case 0x13: /* Z80 */
	  subcpu_memspace[(offset>>1)]=(char)data;
	  break;
        case 0x14: /* PCM */
	  neogeo_pcm_memory[(offset>>1)+(bank<<19)]=(char)data;
	  break;
        default:
         //sprintf(mem_str,"write_upload_word unimplemented zone %x\n",zone);
         //debug_log(mem_str);
			break;
    }
}

static void write_memcard(int offset, int data) {
#ifdef DEBUG_MEMORY
if (trazando) printf("write_memcard(0x%X,0x%X)\n",offset,data); fflush(stdout);
#endif
    data&=0xff;
    neogeo_memorycard[(offset&0x3fff)>>1]=(char)data;

    /* signal that memory card has been written */
    memcard_write=3;
}

