#include "psp.h"
#include "pg.h"
#include "neocd.h"

/* Font support from SDL_gfx : these fonts are very easy to use, so it
   would be a shame not to give them a try ! */

char default_font[20];
static int default_font_loaded = 0;
static char *loaded_font;
static int charWidth, charHeight, charPitch, charSize;

void load_font(char *name, int width, int height) {
  FILE *f;
  int len;
  if (!width || !height) { // load default font
    charWidth = charHeight = charPitch = charSize = 0;
    if (loaded_font) {
      free(loaded_font);
      loaded_font = NULL;
    }
    return;
  }
  if (strchr(name,'/') == 0) {
    char path[256];
    sprintf(path,"%sfonts/%s",launchDir,name);
    f = fopen(path,"rb");
  } else
    f = fopen(name,"rb");

  if (f) {
    fseek(f,0L,SEEK_END);
    len = ftell(f);
    fseek(f,0L,SEEK_SET);

    if (loaded_font)
      free(loaded_font);

    loaded_font = malloc(len);
    default_font_loaded = 0;
    if (loaded_font) {
      fread(loaded_font,len,1,f);
      charWidth = width;
      charHeight = height;
      charPitch = (charWidth+7)/8;
      charSize = charPitch * charHeight;
    } else {
      charWidth = charHeight = charPitch = charSize = 0;
      if (loaded_font) {
	free(loaded_font);
	loaded_font = NULL;
      }
    }
    fclose(f);
  }
}

void pgPutChar_font(unsigned long x,unsigned long y,unsigned long color,unsigned long bgcolor,unsigned char ch,char drawfg,char drawbg,char mag)
{
	char *vptr0;		//pointer to vram
	char *vptr;		//pointer to vram
	const char *cfont;		//pointer to font
	unsigned long cx,cy;
	u8 mask,patt;
	char mx,my;

	if (!charSize)
	  return pgPutChar(x,y,color,bgcolor,ch,drawfg,drawbg,mag);

	cfont=loaded_font+ch*charSize;
	vptr0=pgGetVramAddr(x,y);
	for (cy=0; cy<charHeight; cy++) {
	  for (my=0; my<mag; my++) {
	    vptr=vptr0;
	    mask=0;
	    for (cx=0; cx<charWidth; cx++) {
	      if (!(mask >>= 1)) {
		patt = *cfont++;
		mask = 0x80;
	      }
	      for (mx=0; mx<mag; mx++) {
		if ((patt & mask)!=0) {
		  if (drawfg) *(unsigned short *)vptr=color;
		} else {
		  if (drawbg) *(unsigned short *)vptr=bgcolor;
		}
		vptr+=PIXELSIZE*2;
	      }
	    }
	    vptr0+=LINESIZE*2;
	  }
	}
}

void get_font_dimensions(char *s, int *width, int *height) {
  while ((*s < '0' || *s > '9') && *s)
    s++;
  *width = atoi(s);
  while ((*s >= '0' && *s <= '9')) // find separator
    s++;
  while ((*s < '0' || *s > '9') && *s) // find next number
    s++;
  *height = atoi(s);
}

void load_default_font() {
  if (default_font_loaded)
    return;
  if (*default_font) {
    int width,height;
    char tmp_str[256];
    sprintf(tmp_str,"%sfonts/%s",launchDir,default_font);
    get_font_dimensions(default_font,&width,&height);
    load_font(tmp_str,width,height);
  } else
    load_font("",0,0);
  default_font_loaded = 1;
}

void pgPrint_font(unsigned long x,unsigned long y,unsigned long color,const char *str,int bg)
{
  int x1;
  char tab[2] = {0,0};
  x1=x;

  if (!loaded_font)
    return pgPrint(x,y,color,str,bg);

  while (*str!=0 && x1<SCREEN_WIDTH && y<SCREEN_HEIGHT) {
    pgPutChar_font(x1,y,color,0,*str,1,bg,1);
    tab[0] = *str;
    x1+=charWidth;
    str++;
    if (x1>=SCREEN_WIDTH) {
      x1=x;
      y+=charHeight;
    }
  }
}

void pgPrintSel_font(unsigned long x,unsigned long y,unsigned long color,const char *str, unsigned long xmin, unsigned long xmax)
{
  if (!loaded_font)
    return pgPrintSel(x,y,color,str,xmin,xmax);

  pgFillBoxHalfer(xmin,y-2,xmax,y+charHeight,50);
  pgPrint_font(x,y,color,str,0);
}

int get_font_width() { return (charWidth ? charWidth : 8); }
int get_font_height() { return (charHeight ? charHeight : 8); }

void pgPrintDecTrim_font(int x,int y,short col,unsigned int dec,int bg)
{
    char string[12]/* 0125_DELETE_START ,stringfinal[12] 0125_DELETE_END */;
    int i,a;

    for(i=0;i<11;i++) {
      a = dec % 10;
      dec/=10;
      string[10-i]=0x30+a;
      //        if(dec=0 && a==0) break;
    }
    string[i]=0;
    a=0;
    while ((string[a]=='0')&&(a<i-1)) a++;
    pgPrint_font(x,y,col,string+a,bg);
}

// print to max then trim
void pgPrintMax_font(unsigned long x,unsigned long y,unsigned long color,const char *str,int bg, unsigned long xmax, int align) // 0125_EDITED
{
	int xm=x, i=0, j=0; // 0125_EDITED
	char str2[255]; // 0125_EDITED
	char tab[2] = {0,0}; // 0125_EDITED
	memset(str2, 0, 255);
	xmax-=10;

	if(strlen(str)*get_font_width()>=xmax) {
	  while (*str!=0 && xm<(xmax)) {
	    tab[0]=*str;
	    str2[i]=*str;
	    xm+=strlen(tab)*get_font_width();
	    str++;
	    i++;
	  }
	  if( str[strlen(str)-1] == '/' ) { str2[j]='/'; i--; }
	  for(j=i;j>i-3;j--) str2[j]='.';
	} else {
	  strcpy(str2,str);
	}

	if(align==0) // align left
		pgPrint_font(x,y,color,str2,0);
	else // align right
	  pgPrint_font(SCREEN_WIDTH-x-strlen(str2)*get_font_width(),y,color,str2,0);
}

void pgPrintSelMax_font(unsigned long x,unsigned long y,unsigned long color,const char *str, unsigned long xmin, unsigned long xmax)
{
  if (!loaded_font)
    return pgPrintSelMax(x,y,color,str,xmin,xmax);

  pgFillBoxHalfer(xmin,y-2,xmax,y+charHeight,50);
  pgPrintMax_font(x,y,color,str,0,xmax,0);
}

int strwidth_font(const char *s) {
  if (loaded_font)
    return strlen(s)*charWidth;
  return strwidth(s);
}

void pgPrintCenter_font(unsigned long y,unsigned long color,const char *str,int bg){
	long x=((480-strwidth_font(str))/2);
	pgPrint_font(x,y,color,str,bg);
}
