/***************************************************************************

  2610intf.c

  The YM2610 emulator supports up to 2 chips.
  Each chip has the following connections:
  - Status Read / Control Write A
  - Port Read / Data Write A
  - Control Write B
  - Data Write B

***************************************************************************/


#include <stdio.h>
#include "../neocd.h"
#include "ay8910.h"
#include "fm.h"
#include "2610intf.h"
#include "timer.h"

#define YM2610_NUMBUF 2

extern int neogeo_pcm_size; // neocd.c

#if BUILD_YM2610

/* use FM.C with stream system */

// static int stream;
static const struct YM2610interface *intf;

static void *Timer[MAX_2610][2];

/*------------------------- TM2610 -------------------------------*/
/* IRQ Handler */
static void IRQHandler(int n, int irq)
{
    //printf("IRQ!!!\n");
  // printf("irq %d,%d\n",n,irq);
    neogeo_sound_irq(irq);
}

/* Timer overflow callback from timer.c */
static void timer_callback_2610(int param)
{
	int n=param&0x7f;
	int c=param>>7;

/*	logerror("2610 TimerOver %d\n",c); */
	if (Timer[n][c]) timer_remove(Timer[n][c]);
	Timer[n][c] = 0;
	YM2610TimerOver(n,c);
}

/* TimerHandler from fm.c */
static void TimerHandler(int n,int c,int count,double stepTime)
{
	if( count == 0 )
	{	/* Reset FM Timer */
		if( Timer[n][c] )
		{
/*			logerror("2610 TimerReset %d\n",c); */
	 		timer_remove (Timer[n][c]);
			Timer[n][c] = 0;
		}
	}
	else
	{	/* Start FM Timer */
		double timeSec = (double)count * stepTime;

		if( Timer[n][c] == 0 )
		{
			Timer[n][c] = timer_set (timeSec , (c<<7)|n, timer_callback_2610 );
		}
	}
}

static void FMTimerInit( void )
{
	int i;

	for( i = 0 ; i < MAX_2610 ; i++ )
		Timer[i][0] = Timer[i][1] = 0;
	reset_timers();
}

/* update request from fm.c */
void YM2610UpdateRequest(void)
{
  // streamupdate(100);
}
#if BUILD_YM2610B
int YM2610B_sh_start(const struct YM2610interface *msound)
{
	int i,j;
	int rate = audio_sample_rate;
	char buf[YM2610_NUMBUF][40];
	const char *name[YM2610_NUMBUF];
	int mixed_vol,vol[YM2610_NUMBUF];
	void *pcmbufa[YM2610_NUMBUF],*pcmbufb[YM2610_NUMBUF];
	int  pcmsizea[YM2610_NUMBUF],pcmsizeb[YM2610_NUMBUF];

	intf = msound;
	if( intf->num > MAX_2610 ) return 1;

	if (AY8910_sh_start_ym((const struct AY8910interface *)msound)) return 1;

	/* Timer Handler set */
	FMTimerInit();

	/* stream system initialize */
	for (i = 0;i < intf->num;i++)
	{
		/* stream setup */
		mixed_vol = intf->volumeFM[i];
		/* stream setup */
		for (j = 0 ; j < YM2610_NUMBUF ; j++)
		{
			name[j]=buf[j];
			vol[j] = mixed_vol & 0xffff;
			mixed_vol>>=16;
			sprintf(buf[j],"%s #%d Ch%d","YM2610B",i,j+1);
		}
		/* stream[i] = */ stream_init_multim(YM2610_NUMBUF,name,vol,rate,i,YM2610BUpdateOne);
		/* setup adpcm buffers */
		// pcmbufa[i]  = (void *)(memory_region(msound->pcmroma[i]));
		pcmbufa[i] = (void *) neogeo_pcm_memory;
		// pcmsizea[i] = memory_region_length(msound->pcmroma[i]);
		pcmsizea[i] = neogeo_pcm_size;
		pcmbufb[i]  = NULL; // (void *)(memory_region(msound->pcmromb[i]));
		pcmsizeb[i] = 0; // memory_region_length(msound->pcmromb[i]);
/* 		pcmbufa[i]  = (void *)(memory_region(intf->pcmroma[i])); */
/* 		pcmsizea[i] = memory_region_length(intf->pcmroma[i]); */
/* 		pcmbufb[i]  = (void *)(memory_region(intf->pcmromb[i])); */
/* 		pcmsizeb[i] = memory_region_length(intf->pcmromb[i]); */
	}

	/**** initialize YM2610 ****/
	if (YM2610Init(intf->num,intf->baseclock,rate,
		           pcmbufa,pcmsizea,pcmbufb,pcmsizeb,
		           TimerHandler,IRQHandler) == 0)
		return 0;

	/* error */
	return 1;
}
#endif

int YM2610_sh_start(const struct YM2610interface *msound)
{
	int i,j;
	int rate = audio_sample_rate;
	char buf[YM2610_NUMBUF][40];
	const char *name[YM2610_NUMBUF];
	int mixed_vol,vol[YM2610_NUMBUF];
	void *pcmbufa[YM2610_NUMBUF],*pcmbufb[YM2610_NUMBUF];
	int  pcmsizea[YM2610_NUMBUF],pcmsizeb[YM2610_NUMBUF];

	if( msound->num > MAX_2610 ) return 1;

	intf = msound;
	if (AY8910_sh_start((const struct AY8910interface *)msound)) return 1;

	/* Timer Handler set */
	FMTimerInit();

	/* stream system initialize */
	for (i = 0;i < msound->num;i++)
	{
		/* stream setup */
		mixed_vol = msound->volumeFM[i];
		/* stream setup */
		for (j = 0 ; j < YM2610_NUMBUF ; j++)
		{
			name[j]=buf[j];
			vol[j] = mixed_vol & 0xffff;
			mixed_vol>>=16;
			sprintf(buf[j],"%s #%d Ch%d","YM2610",i,j+1);
		}
		/* stream[i] = */ stream_init_multim(YM2610_NUMBUF,name,vol,rate,i,YM2610UpdateOne);
		/* setup adpcm buffers */
		// pcmbufa[i]  = (void *)(memory_region(msound->pcmroma[i]));
		pcmbufa[i] = (void *) neogeo_pcm_memory;
		// pcmsizea[i] = memory_region_length(msound->pcmroma[i]);
		pcmsizea[i] = neogeo_pcm_size;
		pcmbufb[i]  = (void *) neogeo_pcm_memory; // NULL; // (void *)(memory_region(msound->pcmromb[i]));
		pcmsizeb[i] = neogeo_pcm_size; // memory_region_length(msound->pcmromb[i]);
	}

	/**** initialize YM2610 ****/
	if (YM2610Init(msound->num,msound->baseclock,rate,
		           pcmbufa,pcmsizea,pcmbufb,pcmsizeb,
		           TimerHandler,IRQHandler) == 0)
		return 0;

	/* error */
	return 1;
}
#if 0
int YM2610_sh_start(const struct YM2610interface *msound)
{
    int j;
    int rate = SAMPLE_RATE;
    char buf[YM2610_NUMBUF][40];
    void *pcmbufa[YM2610_NUMBUF],*pcmbufb[YM2610_NUMBUF];
    int  pcmsizea[YM2610_NUMBUF],pcmsizeb[YM2610_NUMBUF];

    if (AY8910_sh_start((const struct AY8910interface *)msound)) return 1;

    /* Timer Handler set */
    FMTimerInit();
    for (j = 0; j < YM2610_NUMBUF; j++) {
	buf[j][0] = 0;
    }
    stream = stream_init_multi(YM2610_NUMBUF, 0, YM2610UpdateOne);
    pcmbufa = (void *) neogeo_pcm_memory;
    pcmbufb = NULL; /*(void *) memory.sound2;*/

    //}

  /**** initialize YM2610 ****/
    /*
       if (YM2610Init(8000000,rate,
       pcmbufa,pcmsizea,pcmbufb,pcmsizeb,
       TimerHandler,IRQHandler) == 0)
     */
    if (YM2610Init(msound->num,msound->baseclock,rate,
		   pcmbufa, neogeo_pcm_size, pcmbufb, 0,
		   TimerHandler, IRQHandler) == 0)
	return 0;

    /* error */
    return 1;
}
#endif

/************************************************/
/* Sound Hardware Stop				*/
/************************************************/
void YM2610_sh_stop(void)
{
    YM2610Shutdown();
}

/* reset */
void YM2610_sh_reset(void)
{
	int i;

	for (i = 0;i < intf->num;i++)
		YM2610ResetChip(i);
}

/************************************************/
/* Status Read for YM2610 - Chip 0		*/
/************************************************/
u32 YM2610_status_port_0_A_r(u32 offset)
{
    return YM2610Read(0,0);
}

u32 YM2610_status_port_0_B_r(u32 offset)
{
    return YM2610Read(0,2);
}

/************************************************/
/* Port Read for YM2610 - Chip 0		*/
/************************************************/
u32 YM2610_read_port_0_r(u32 offset)
{
    return YM2610Read(0,1);
}


/************************************************/
/* Control Write for YM2610 - Chip 0		*/
/* Consists of 2 addresses			*/
/************************************************/
void YM2610_control_port_0_A_w(u32 offset, u32 data)
{
    YM2610Write(0, 0,data);
}

void YM2610_control_port_0_B_w(u32 offset, u32 data)
{
    YM2610Write(0,2, data);
}

/************************************************/
/* Data Write for YM2610 - Chip 0		*/
/* Consists of 2 addresses			*/
/************************************************/
void YM2610_data_port_0_A_w(u32 offset, u32 data)
{
    YM2610Write(0,1, data);
}

void YM2610_data_port_0_B_w(u32 offset, u32 data)
{
    YM2610Write(0,3, data);
}

#endif
